from collections import deque
from math import ceil

import gym
from fastai.basic_train import *
from fastai.torch_core import *

from fast_rl.core.data_structures import SumTree


class ExplorationStrategy:
	def __init__(self, explore: bool = True): self.explore=explore
	def update(self, max_episodes, explore, **kwargs): self.explore=explore
	def perturb(self, action, action_space) -> np.ndarray:
		"""
		Base method just returns the action. Subclass, and change to return randomly / augmented actions.

		Should use `do_exploration` field. It is recommended that when you subclass / overload, you allow this field
		to completely bypass these actions.

		Args:
			action (np.array): 			Action input as a regular numpy array.
			action_space (gym.Space): 	The original gym space. Should contain information on the action type, and
										possible convenience methods for random action selection.
		"""
		_=action_space
		return action


class GreedyEpsilon(ExplorationStrategy):
	def __init__(self, epsilon_start, epsilon_end, decay, start_episode=0, end_episode=0, **kwargs):
		super().__init__(**kwargs)
		self.end_episode=end_episode
		self.start_episode=start_episode
		self.decay=decay
		self.e_end=epsilon_end
		self.e_start=epsilon_start
		self.epsilon=self.e_start
		self.steps=0

	def perturb(self, action, action_space: gym.Space):
		return action_space.sample() if np.random.random()<self.epsilon and self.explore else action

	def update(self, episode, end_episode=0, **kwargs):
		super(GreedyEpsilon, self).update(**kwargs)
		if self.explore:
			self.end_episode=end_episode
			self.epsilon=self.e_end+(self.e_start-self.e_end)*math.exp(-1.*(self.steps*self.decay))
			self.steps+=1


class OrnsteinUhlenbeck(GreedyEpsilon):
	def __init__(self, size, mu=0., theta=0.15, sigma=0.2, **kwargs):
		"""

		References:
			[1] From https://math.stackexchange.com/questions/1287634/implementing-ornstein-uhlenbeck-in-matlab
			[2] Cumulatively based on

		Args:
			epsilon_start:
			epsilon_end:
			decay:
			**kwargs:
		"""
		super().__init__(**kwargs)
		self.sigma=sigma
		self.theta=theta
		self.mu=mu
		self.x=np.ones(size)

	def perturb(self, action, action_space):
		dx=np.zeros(self.x.shape)
		if self.explore:
			dx=self.theta*(self.mu-self.x)+self.sigma*np.array([np.random.normal() for _ in range(len(self.x))])

		self.x+=dx
		return self.epsilon*self.x+action


class Experience:
	def __init__(self, memory_size, reduce_ram=False):
		self.reduce_ram=reduce_ram
		self.max_size=memory_size
		self.callbacks=[]

	@property
	def memory(self): return None
	def sample(self, **kwargs): pass
	def update(self, item, **kwargs): item.to(device=defaults.device)
	def refresh(self, **kwargs): pass


class ExperienceReplay(Experience):
	def __init__(self, memory_size, **kwargs):
		r"""
		Basic store-er of s space transitions for training agents.

		References:
			[1] Mnih, Volodymyr, et al. "Playing atari with deep reinforcement learning."
			arXiv preprint arXiv:1312.5602 (2013).

		Args:
			memory_size (int): Max N samples to store
		"""
		super().__init__(memory_size, **kwargs)
		self.max_size=memory_size
		self._memory=deque(maxlen=memory_size)

	@property
	def memory(self):
		return self._memory

	def __len__(self):
		return len(self._memory)

	def sample(self, batch, **kwargs):
		if len(self._memory)<batch: return self._memory
		return random.sample(self.memory, batch)

	def update(self, item, **kwargs):
		item=deepcopy(item)
		super().update(item, **kwargs)
		if self.reduce_ram: item.clean()
		self._memory.append(item)


class PriorityExperienceReplayCallback(LearnerCallback):
	def on_train_begin(self, **kwargs):
		self.learn.model.loss_func=partial(self.learn.model.memory.handle_loss, loss_fn=self.learn.model.loss_func)


class PriorityExperienceReplay(Experience):

	def handle_loss(self, y, y_hat, loss_fn):
		return (loss_fn(y, y_hat)*torch.from_numpy(self.p_weights).to(device=defaults.device).float()).mean()

	def __init__(self, memory_size, batch_size=64, epsilon=0.01, alpha=0.6, beta=0.4, b_inc=-0.001, **kwargs):
		"""
		Prioritizes sampling based on samples requiring the most learning.

		References:
			[1] Schaul, Tom, et al. "Prioritized experience replay." arXiv preprint arXiv:1511.05952 (2015).

		Args:
			batch_size (int): Size of sample, and thus size of expected index update.
			alpha (float): Changes the sampling behavior 1 (non-uniform) -> 0 (uniform)
			epsilon (float): Keeps the probabilities of items from being 0
			memory_size (int): Max N samples to store
		"""
		super().__init__(memory_size, **kwargs)
		self.batch_size=batch_size
		self.alpha=alpha
		self.beta=beta
		self.b_inc=b_inc
		self.p_weights=None  # np.zeros(self.batch_size, dtype=float)
		self.epsilon=epsilon
		self.tree=SumTree(self.max_size)
		self.callbacks=[PriorityExperienceReplayCallback]
		# When sampled, store the sample indices for refresh.
		self._indices=None  # np.zeros(self.batch_size, dtype=int)

	@property
	def memory(self):
		return self.tree.data

	def __len__(self):
		return self.tree.n_entries

	def refresh(self, post_optimize, **kwargs):
		if post_optimize is not None:
			self.tree.update(self._indices.astype(int), np.abs(post_optimize['td_error'])+self.epsilon)

	def sample(self, batch, **kwargs):
		self.beta=np.min([1., self.beta+self.b_inc])
		ranges=np.linspace(0, ceil(self.tree.total()/batch), num=batch+1)
		uniform_ranges=[np.random.uniform(ranges[i], ranges[i+1]) for i in range(len(ranges)-1)]
		try:
			self._indices, weights, samples=self.tree.batch_get(uniform_ranges)
		except ValueError:
			warn('Too few values to unpack. Your batch size is too small, when PER queries tree, all 0 values get'
				 ' ignored. We will retry until we can return at least one sample.')
			samples=self.sample(batch)
			return samples

		self.p_weights=self.tree.anneal_weights(weights, self.beta)
		return samples

	def update(self, item, **kwargs):
		"""
		Updates the tree of PER.

		Assigns maximal priority per [1] Alg:1, thus guaranteeing that sample being visited once.

		Args:
			item:

		Returns:

		"""
		item=deepcopy(item)
		super().update(item, **kwargs)
		maximal_priority=self.alpha
		if self.reduce_ram: item.clean()
		self.tree.add(np.abs(maximal_priority)+self.epsilon, item)


class HindsightExperienceReplay(Experience):
	def __init__(self, memory_size):
		"""

		References:
			[1] Andrychowicz, Marcin, et al. "Hindsight experience replay."
			Advances in Neural Information Processing Systems. 2017.

		Args:
			memory_size:
		"""
		super().__init__(memory_size)



r"""`fast_rl.layers` provides essential functions to building and modifying `model` architectures"""
from math import ceil

from fastai.torch_core import *
from fastai.tabular import TabularModel


def init_cnn(mod: Any):
	r""" Utility for initializing cnn Modules. """
	if getattr(mod, 'bias', None) is not None: nn.init.constant_(mod.bias, 0)
	if isinstance(mod, (nn.Conv2d, nn.Linear)): nn.init.kaiming_normal_(mod.weight)
	for sub_mod in mod.children(): init_cnn(sub_mod)


def ks_stride(ks, stride, w, h, n_blocks, kern_proportion=.1, stride_proportion=0.3):
	r""" Utility for determing the the kernel size and stride. """
	kernels, strides, max_dim = [], [], max((w, h))
	for i in range(len(n_blocks)):
		kernels.append(max_dim * kern_proportion)
		strides.append(kernels[-1] * stride_proportion)
		max_dim = (max_dim - kernels[-1]) / strides[-1]
		assert max_dim > 1

	return ifnone(ks, map(ceil, kernels)), ifnone(stride, map(ceil, strides))


class Flatten(nn.Module):
	def forward(self, y): return y.view(y.size(0), -1)


class FakeBatchNorm(Module):
	r""" If we want all the batch norm layers gone, then we will replace the tabular batch norm with this. """
	def forward(self, xi: Tensor, *args): return xi


def conv_bn_lrelu(ni: int, nf: int, ks: int = 3, stride: int = 1, pad=True, bn=True) -> nn.Sequential:
	r""" Create a sequence Conv2d->BatchNorm2d->LeakyReLu layer. (from darknet.py). Allows excluding BatchNorm2d Layer."""
	return nn.Sequential(
		nn.Conv2d(ni, nf, kernel_size=ks, bias=False, stride=stride, padding=(ks // 2) if pad else 0),
		nn.BatchNorm2d(nf) if bn else FakeBatchNorm(),
		nn.LeakyReLU(negative_slope=0.1, inplace=True))


class ChannelTranspose(Module):
	r""" Runtime image input channel changing. Useful for handling different image channel outputs from different envs. """
	def forward(self, xi: Tensor):
		return xi.transpose(3, 1).transpose(3, 2)


class StateActionSplitter(Module):
	r""" `Actor / Critic` models require breaking the state and action into 2 streams. """

	def forward(self, s_a_tuple: Tuple[Tensor]):
		r""" Returns tensors as -> (State Tensor, Action Tensor) """
		return s_a_tuple[0], s_a_tuple[1]


class StateActionPassThrough(nn.Module):
	r""" Passes action input untouched, but runs the state tensors through a sub module. """
	def __init__(self, layers):
		super().__init__()
		self.layers = layers

	def forward(self, state_action):
		return self.layers(state_action[0]), state_action[1]


class TabularEmbedWrapper(Module):
	r""" Basic `TabularModel` compatibility wrapper. Typically, state inputs will be either categorical or continuous. """
	def __init__(self, tabular_model: TabularModel):
		super().__init__()
		self.tabular_model = tabular_model

	def forward(self, xi: Tensor, *args):
		return self.tabular_model(xi, xi)


class CriticTabularEmbedWrapper(Module):
	r""" Similar to `TabularEmbedWrapper` but assumes input is state / action and requires concatenation. """
	def __init__(self, tabular_model: TabularModel, exclude_cat):
		super().__init__()
		self.tabular_model = tabular_model
		self.exclude_cat = exclude_cat

	def forward(self, args):
		if not self.exclude_cat:
			return self.tabular_model(*args)
		else:
			return self.tabular_model(0, torch.cat(args, axis=1))



from functools import partial
from itertools import product

import pytest
from fastai.basic_train import torch, DatasetType
from fastai.core import ifnone

from fast_rl.agents.ddpg import create_ddpg_model, ddpg_learner, DDPGLearner
from fast_rl.agents.ddpg_models import DDPGModule
from fast_rl.core.agent_core import ExperienceReplay, PriorityExperienceReplay, OrnsteinUhlenbeck
from fast_rl.core.data_block import FEED_TYPE_STATE, MDPDataBunch, ResolutionWrapper
from fast_rl.core.metrics import RewardMetric, EpsilonMetric
from fast_rl.core.train import GroupAgentInterpretation, AgentInterpretation

p_model=[DDPGModule]
p_exp=[ExperienceReplay, PriorityExperienceReplay]
p_format=[FEED_TYPE_STATE]  # , FEED_TYPE_IMAGE]
p_full_format=[FEED_TYPE_STATE]
p_envs=['Pendulum-v0']

config_env_expectations={
	'Pendulum-v0': {'action_shape': (1, 1), 'state_shape': (1, 3)}
}


def trained_learner(model_cls, env, s_format, experience, bs=64,layers=None,render='rgb_array', memory_size=1000000,
					decay=0.0001,lr=None,actor_lr=None,epochs=450,opt=torch.optim.RMSprop, **kwargs):
	lr,actor_lr=ifnone(lr,1e-3),ifnone(actor_lr,1e-4)
	data=MDPDataBunch.from_env(env,render=render,bs=bs,add_valid=False,keep_env_open=False,feed_type=s_format,
							   memory_management_strategy='k_partitions_top',k=3,**kwargs)
	exploration_method=OrnsteinUhlenbeck(size=data.action.taken_action.shape,epsilon_start=1,epsilon_end=0.1,
										 decay=decay)
	memory=experience(memory_size=memory_size, reduce_ram=True)
	model=create_ddpg_model(data=data,base_arch=model_cls,lr=lr,actor_lr=actor_lr,layers=layers,opt=opt)
	learner=ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method,
						 callback_fns=[RewardMetric, EpsilonMetric])
	learner.fit(epochs)
	return learner

def check_shape(env,data,s_format):
	assert config_env_expectations[env]['action_shape']==(1, data.action.taken_action.shape[1])
	if s_format==FEED_TYPE_STATE:
		assert config_env_expectations[env]['state_shape']==data.state.s.shape

def learner2gif(lnr:DDPGLearner,s_format,group_interp:GroupAgentInterpretation,name:str,extra:str):
	meta=f'{lnr.memory.__class__.__name__}_{"FEED_TYPE_STATE" if s_format==FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
	interp=AgentInterpretation(lnr, ds_type=DatasetType.Train)
	interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
	group_interp.add_interpretation(interp)
	group_interp.to_pickle(f'../docs_src/data/{name}_{lnr.model.name.lower()}/', f'{lnr.model.name.lower()}_{meta}')
	[g.write(f'../res/run_gifs/{name}_{extra}') for g in interp.generate_gif()]


@pytest.mark.parametrize(["model_cls", "s_format", "env"], list(product(p_model, p_format, p_envs)))
def test_ddpg_create_ddpg_model(model_cls, s_format, env):
	data=MDPDataBunch.from_env(env, render='rgb_array', bs=32, add_valid=False, feed_type=s_format)
	model=create_ddpg_model(data, model_cls)
	model.eval()
	model(data.state.s.float())
	check_shape(env,data,s_format)
	data.close()


@pytest.mark.parametrize(["model_cls", "s_format", "mem", "env"], list(product(p_model, p_format, p_exp, p_envs)))
def test_ddpg_ddpglearner(model_cls, s_format, mem, env):
	data=MDPDataBunch.from_env(env, render='rgb_array', bs=32, add_valid=False, feed_type=s_format)
	model=create_ddpg_model(data, model_cls)
	memory=mem(memory_size=1000, reduce_ram=True)
	exploration_method=OrnsteinUhlenbeck(size=data.action.taken_action.shape, epsilon_start=1, epsilon_end=0.1,
		decay=0.001)
	ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
	check_shape(env,data,s_format)
	data.close()


@pytest.mark.parametrize(["model_cls", "s_format", "mem", "env"], list(product(p_model, p_format, p_exp, p_envs)))
def test_ddpg_fit(model_cls, s_format, mem, env):
	learner=trained_learner(env=env, bs=10,opt=torch.optim.RMSprop,model_cls=model_cls,layers=[20, 20],memory_size=100,
							max_steps=20,render='rgb_array',decay=0.001,s_format=s_format,experience=mem,epochs=2)

	check_shape(env,learner.data,s_format)
	del learner


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'],
	list(product(p_model, p_format, p_exp)))
def test_ddpg_models_pendulum(model_cls, s_format, experience):
	group_interp=GroupAgentInterpretation()
	extra_s=f'{experience.__name__}_{model_cls.__name__}_{s_format}'
	for i in range(5):
		print('\n')
		learner=trained_learner(model_cls,'Pendulum-v0',s_format,experience,decay=0.0001,render='rgb_array')
		learner2gif(learner,s_format,group_interp,'pendulum',extra_s)
		del learner


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'],
	list(product(p_model, p_format, p_exp)))
def test_ddpg_models_acrobot(model_cls, s_format, experience):
	group_interp=GroupAgentInterpretation()
	extra_s=f'{experience.__name__}_{model_cls.__name__}_{s_format}'
	for i in range(5):
		print('\n')
		learner=trained_learner(model_cls,'Acrobot-v1',s_format,experience,decay=0.0001,render='rgb_array')
		learner2gif(learner,s_format,group_interp,'acrobot',extra_s)
		del learner


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'],
	list(product(p_model, p_full_format, p_exp)))
def test_ddpg_models_mountain_car_continuous(model_cls, s_format, experience):
	group_interp=GroupAgentInterpretation()
	for i in range(5):
		print('\n')
		data=MDPDataBunch.from_env('MountainCarContinuous-v0', render='rgb_array', bs=40, add_valid=False, keep_env_open=False,
			feed_type=s_format, memory_management_strategy='k_partitions_top', k=3, res_wrap=partial(ResolutionWrapper, w_step=2, h_step=2))
		exploration_method=OrnsteinUhlenbeck(size=data.action.taken_action.shape, epsilon_start=1, epsilon_end=0.1,
			decay=0.0001)
		memory=experience(memory_size=1000000, reduce_ram=True)
		model=create_ddpg_model(data=data, base_arch=model_cls)
		learner=ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method,
			callback_fns=[RewardMetric, EpsilonMetric])
		learner.fit(450)

		meta=f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format==FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
		interp=AgentInterpretation(learner, ds_type=DatasetType.Train)
		interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
		group_interp.add_interpretation(interp)
		group_interp.to_pickle(f'../docs_src/data/mountaincarcontinuous_{model.name.lower()}/',
			f'{model.name.lower()}_{meta}')
		[g.write('../res/run_gifs/mountaincarcontinuous') for g in interp.generate_gif()]
		data.close()
		del learner
		del model
		del data


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'],
	list(product(p_model, p_full_format, p_exp)))
def test_ddpg_models_reach(model_cls, s_format, experience):
	group_interp=GroupAgentInterpretation()
	for i in range(5):
		print('\n')
		data=MDPDataBunch.from_env('ReacherPyBulletEnv-v0', render='rgb_array', bs=40, add_valid=False, keep_env_open=False, feed_type=s_format,
			memory_management_strategy='k_partitions_top', k=3, res_wrap=partial(ResolutionWrapper, w_step=2, h_step=2))
		exploration_method=OrnsteinUhlenbeck(size=data.action.taken_action.shape, epsilon_start=1, epsilon_end=0.1,
			decay=0.00001)
		memory=experience(memory_size=1000000, reduce_ram=True)
		model=create_ddpg_model(data=data, base_arch=model_cls)
		learner=ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method,
			callback_fns=[RewardMetric, EpsilonMetric])
		learner.fit(450)

		meta=f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format==FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
		interp=AgentInterpretation(learner, ds_type=DatasetType.Train)
		interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
		group_interp.add_interpretation(interp)
		group_interp.to_pickle(f'../docs_src/data/reacher_{model.name.lower()}/',
			f'{model.name.lower()}_{meta}')
		[g.write('../res/run_gifs/reacher') for g in interp.generate_gif()]
		data.close()
		del learner
		del model
		del data


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'],
	list(product(p_model, p_full_format, p_exp)))
def test_ddpg_models_walker(model_cls, s_format, experience):
	group_interp=GroupAgentInterpretation()
	extra_s=f'{experience.__name__}_{model_cls.__name__}_{s_format}'
	for i in range(5):
		print('\n')
		# data=MDPDataBunch.from_env('Walker2DPyBulletEnv-v0', render='human', bs=64, add_valid=False, keep_env_open=False,
		# 	feed_type=s_format, memory_management_strategy='k_partitions_top', k=3)
		# exploration_method=OrnsteinUhlenbeck(size=data.action.taken_action.shape, epsilon_start=1, epsilon_end=0.1,
		# 	decay=0.0001)
		# memory=experience(memory_size=1000000, reduce_ram=True)
		# model=create_ddpg_model(data=data, base_arch=model_cls)
		# learner=ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method,
		# 	callback_fns=[RewardMetric, EpsilonMetric])
		# learner.fit(1500)
		learner=trained_learner(model_cls,'Walker2DPyBulletEnv-v0',s_format,experience,decay=0.0001,render='rgb_array')
		learner2gif(learner,s_format,group_interp,'walker2d',extra_s)

		# meta=f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format==FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
		# interp=AgentInterpretation(learner, ds_type=DatasetType.Train)
		# interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
		# group_interp.add_interpretation(interp)
		# group_interp.to_pickle(f'../docs_src/data/walker2d_{model.name.lower()}/',
		# 	f'{model.name.lower()}_{meta}')
		# [g.write('../res/run_gifs/walker2d') for g in interp.generate_gif()]
		# data.close()
		del learner
		# del model
		# del data


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'],
	list(product(p_model, p_full_format, p_exp)))
def test_ddpg_models_ant(model_cls, s_format, experience):
	group_interp=GroupAgentInterpretation()
	extra_s = f'{experience.__name__}_{model_cls.__name__}_{s_format}'
	for i in range(5):
		print('\n')
		# data=MDPDataBunch.from_env('AntPyBulletEnv-v0', render='human', bs=64, add_valid=False, keep_env_open=False,
		# 	feed_type=s_format, memory_management_strategy='k_partitions_top', k=3)
		# exploration_method=OrnsteinUhlenbeck(size=data.action.taken_action.shape, epsilon_start=1, epsilon_end=0.1,
		# 	decay=0.00001)
		# memory=experience(memory_size=1000000, reduce_ram=True)
		# model=create_ddpg_model(data=data, base_arch=model_cls, lr=1e-3, actor_lr=1e-4)
		# learner=ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method,
		# 	opt_func=torch.optim.Adam, callback_fns=[RewardMetric, EpsilonMetric])
		# learner.fit(4)
		learner=trained_learner(model_cls,'AntPyBulletEnv-v0',s_format,experience,decay=0.0001,render='rgb_array',epochs=1000)
		learner2gif(learner,s_format,group_interp,'ant',extra_s)
		# meta=f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format==FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
		# interp=AgentInterpretation(learner, ds_type=DatasetType.Train)
		# interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
		# group_interp.add_interpretation(interp)
		# group_interp.to_pickle(f'../docs_src/data/ant_{model.name.lower()}/',
		# 	f'{model.name.lower()}_{meta}')
		# [g.write('../res/run_gifs/ant', frame_skip=5) for g in interp.generate_gif()]
		del learner
		# del model
		# del data


@pytest.mark.usefixtures('skip_performance_check')
@pytest.mark.parametrize(["model_cls", "s_format", 'experience'],
	list(product(p_model, p_full_format, p_exp)))
def test_ddpg_models_halfcheetah(model_cls, s_format, experience):
	group_interp=GroupAgentInterpretation()
	for i in range(5):
		print('\n')
		data=MDPDataBunch.from_env('HalfCheetahPyBulletEnv-v0', render='rgb_array', bs=64, add_valid=False, keep_env_open=False,
			feed_type=s_format, memory_management_strategy='k_partitions_top', k=3, res_wrap=partial(ResolutionWrapper, w_step=2, h_step=2))
		exploration_method=OrnsteinUhlenbeck(size=data.action.taken_action.shape, epsilon_start=1, epsilon_end=0.1,
			decay=0.000001)
		memory=experience(memory_size=1000000, reduce_ram=True)
		model=create_ddpg_model(data=data, base_arch=model_cls)
		learner=ddpg_learner(data=data, model=model, memory=memory, exploration_method=exploration_method,
			callback_fns=[RewardMetric, EpsilonMetric])
		learner.fit(1000)

		meta=f'{experience.__name__}_{"FEED_TYPE_STATE" if s_format==FEED_TYPE_STATE else "FEED_TYPE_IMAGE"}'
		interp=AgentInterpretation(learner, ds_type=DatasetType.Train)
		interp.plot_rewards(cumulative=True, per_episode=True, group_name=meta)
		group_interp.add_interpretation(interp)
		group_interp.to_pickle(f'../docs_src/data/halfcheetah_{model.name.lower()}/',
			f'{model.name.lower()}_{meta}')
		[g.write('../res/run_gifs/halfcheetah') for g in interp.generate_gif()]
		del learner
		del model
		del data



from multiprocessing.pool import Pool

from fastai.basic_train import Learner, load_callback
from fastai.torch_core import *

from fast_rl.core.data_block import MDPDataBunch


class WrapperLossFunc(object):
	def __init__(self, learn):
		self.learn = learn

	def __call__(self, *args, **kwargs):
		return self.learn.model.loss



def load_learner(path:PathOrStr, file:PathLikeOrBinaryStream='export.pkl', **db_kwargs):
	r""" Similar to fastai `load_learner`, handles load_state for data differently. """
	source = Path(path)/file if is_pathlike(file) else file
	state = torch.load(source, map_location='cpu') if defaults.device == torch.device('cpu') else torch.load(source)
	model = state.pop('model')
	data = MDPDataBunch.load_state(path, state.pop('data'))
	# if test is not None: src.add_test(test)
	# data = src.databunch(**db_kwargs)
	cb_state = state.pop('cb_state')
	clas_func = state.pop('cls')
	res = clas_func(data, model, **state)
	res.callback_fns = state['callback_fns'] #to avoid duplicates
	res.callbacks = [load_callback(c,s, res) for c,s in cb_state.items()]
	return res


class AgentLearner(Learner):

	def __init__(self, data, loss_func=None, callback_fns=None, opt=torch.optim.Adam, **kwargs):
		super().__init__(data=data, callback_fns=ifnone(callback_fns, []) + data.callback, **kwargs)
		self.model.loss_func = ifnone(loss_func, F.mse_loss)
		self.model.set_opt(opt)
		self.loss_func = None
		self.trainers = None
		self._loss_func = WrapperLossFunc(self)

	@property
	def warming_up(self):
		return self.data.bs > len(self.data.x)

	def init_loss_func(self):
		r"""
		 Initializes the loss function wrapper for logging loss.

		 Since most RL models have a period of warming up such as filling tree buffers, we cannot log any loss.
		 By default, the learner will have a `None` loss function, and so the fit function will not try to log that
		 loss.
		 """
		self.loss_func = WrapperLossFunc(self)

	def export(self, file:PathLikeOrBinaryStream='export.pkl', destroy=False, pickle_data=False):
		"Export the state of the `Learner` in `self.path/file`. `file` can be file-like (file or buffer)"
		if rank_distrib(): return # don't save if slave proc
		# For now we exclude the 'loss_func' since it is pointing to a model loss.
		args = ['opt_func', 'metrics', 'true_wd', 'bn_wd', 'wd', 'train_bn', 'model_dir', 'callback_fns', 'memory',
				'exploration_method', 'trainers']
		state = {a:getattr(self,a) for a in args}
		state['cb_state'] = {cb.__class__:cb.get_state() for cb in self.callbacks}
		#layer_groups -> need to find a way
		#TO SEE: do we save model structure and weights separately?
		with ModelOnCPU(self.model) as m:
			m.opt = None
			state['model'] = m
			xtra = dict(normalize=self.data.norm.keywords) if getattr(self.data, 'norm', False) else {}
			state['data'] = self.data.train_ds.get_state(**xtra) if self.data.valid_dl is None else self.data.valid_ds.get_state(**xtra)
			state['data']['add_valid'] = not self.data.empty_val
			if pickle_data: self.data.to_pickle(self.data.path)
			state['cls'] = self.__class__
			try_save(state, self.path, file)
		if destroy: self.destroy()

	def interpret_q(self, xi):
		raise NotImplemented


class PipeLine(object):
	def __init__(self, n_threads, pipe_line_function):
		warn(Warning('Currently not super useful. Seems to have issues with running a single env in multiple threads.'))
		self.pipe_line_function = pipe_line_function
		self.n_threads = n_threads
		self.pool = Pool(self.n_threads)

	def start(self, n_runs):
		return self.pool.map(self.pipe_line_function,  range(n_runs))


import os
from functools import partial
from itertools import product

import gym
import pytest
import numpy as np
import torch
from fastai.basic_train import ItemLists

from fast_rl.agents.dqn import create_dqn_model, dqn_learner
from fast_rl.agents.dqn_models import DQNModule
from fast_rl.core.agent_core import GreedyEpsilon, ExperienceReplay
from fast_rl.core.data_block import MDPDataBunch, ResolutionWrapper, FEED_TYPE_IMAGE
from fast_rl.core.metrics import RewardMetric, EpsilonMetric


def validate_item_list(item_list: ItemLists):
    # Check items
    for i, item in enumerate(item_list.items):
        if item.done: assert not item_list.items[
            i - 1].done, f'The dataset has duplicate "done\'s" that are consecutive.'
        assert item.state.s is not None, f'The item: {item}\'s state is None'
        assert item.state.s_prime is not None, f'The item: {item}\'s state prime is None'


@pytest.mark.parametrize(["memory_strategy", "k"], list(product(['k_top', 'k_partitions_top'], [1, 3, 5])))
def test_dataset_memory_manager(memory_strategy, k):
    data = MDPDataBunch.from_env('CartPole-v0', render='rgb_array', bs=5, max_steps=20, add_valid=False,
                                 memory_management_strategy=memory_strategy, k=k)
    model = create_dqn_model(data, DQNModule, opt=torch.optim.RMSprop, lr=0.1)
    memory = ExperienceReplay(memory_size=1000, reduce_ram=True)
    exploration_method = GreedyEpsilon(epsilon_start=1, epsilon_end=0.1, decay=0.001)
    learner = dqn_learner(data=data, model=model, memory=memory, exploration_method=exploration_method,
                          callback_fns=[RewardMetric, EpsilonMetric])
    learner.fit(10)

    data_info = {episode: data.train_ds.x.info[episode] for episode in data.train_ds.x.info if episode != -1}
    full_episodes = [episode for episode in data_info if not data_info[episode][1]]

    assert sum([not _[1] for _ in data_info.values()]) == k, 'There should be k episodes but there is not.'
    if memory_strategy.__contains__('top') and not memory_strategy.__contains__('both'):
        assert (np.argmax([_[0] for _ in data_info.values()])) in full_episodes


def test_databunch_to_pickle():
    data = MDPDataBunch.from_env('CartPole-v0', render='rgb_array', bs=5, max_steps=20, add_valid=False,
                                 memory_management_strategy='k_partitions_top', k=3)
    model = create_dqn_model(data, DQNModule, opt=torch.optim.RMSprop, lr=0.1)
    memory = ExperienceReplay(memory_size=1000, reduce_ram=True)
    exploration_method = GreedyEpsilon(epsilon_start=1, epsilon_end=0.1, decay=0.001)
    learner = dqn_learner(data=data, model=model, memory=memory, exploration_method=exploration_method,
                          callback_fns=[RewardMetric, EpsilonMetric])
    learner.fit(10)
    data.to_pickle('./data/cartpole_10_epoch')
    MDPDataBunch.from_pickle(env_name='CartPole-v0', path='./data/cartpole_10_epoch')


def test_resolution_wrapper():
    data = MDPDataBunch.from_env('CartPole-v0', render='rgb_array', bs=5, max_steps=10, add_valid=False,
                                memory_management_strategy='k_top', k=1, feed_type=FEED_TYPE_IMAGE,
                                res_wrap=partial(ResolutionWrapper, w_step=2, h_step=2))
    model = create_dqn_model(data, DQNModule, opt=torch.optim.RMSprop, lr=0.1,channels=[32,32,32],ks=[5,5,5],stride=[2,2,2])
    memory = ExperienceReplay(memory_size=1000, reduce_ram=True)
    exploration_method = GreedyEpsilon(epsilon_start=1, epsilon_end=0.1, decay=0.001)
    learner = dqn_learner(data=data, model=model, memory=memory, exploration_method=exploration_method,
                          callback_fns=[RewardMetric, EpsilonMetric])
    learner.fit(2)
    temp = gym.make('CartPole-v0')
    temp.reset()
    original_shape = temp.render(mode='rgb_array').shape
    assert data.env.render(mode='rgb_array').shape == (original_shape[0] // 2, original_shape[1] // 2, 3)



import os



from fast_rl.agents.dqn import create_dqn_model, FixedTargetDQNModule, dqn_learner
from fast_rl.core.agent_core import ExperienceReplay, torch, GreedyEpsilon
from fast_rl.core.basic_train import load_learner
from fast_rl.core.data_block import MDPDataBunch


def test_fit():
	data=MDPDataBunch.from_env('CartPole-v0', render='rgb_array', bs=5, max_steps=20, add_valid=False)
	model=create_dqn_model(data, FixedTargetDQNModule, opt=torch.optim.RMSprop)
	memory=ExperienceReplay(memory_size=1000, reduce_ram=True)
	exploration_method=GreedyEpsilon(epsilon_start=1, epsilon_end=0.1, decay=0.001)
	learner=dqn_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
	learner.fit(2)
	learner.fit(2)
	learner.fit(2)

	assert len(data.x.info)==6
	assert 0 in data.x.info
	assert 5 in data.x.info


def test_to_pickle():
	data=MDPDataBunch.from_env('CartPole-v0', render='rgb_array', bs=5, max_steps=20, add_valid=False)
	model=create_dqn_model(data, FixedTargetDQNModule, opt=torch.optim.RMSprop)
	memory=ExperienceReplay(memory_size=1000, reduce_ram=True)
	exploration_method=GreedyEpsilon(epsilon_start=1, epsilon_end=0.1, decay=0.001)
	learner=dqn_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
	learner.fit(2)

	assert len(data.x.info)==2
	assert 0 in data.x.info
	assert 1 in data.x.info

	data.to_pickle('./data/test_to_pickle')
	assert os.path.exists('./data/test_to_pickle_CartPole-v0')


def test_from_pickle():
	data=MDPDataBunch.from_pickle('./data/test_to_pickle_CartPole-v0')
	model=create_dqn_model(data, FixedTargetDQNModule, opt=torch.optim.RMSprop)
	memory=ExperienceReplay(memory_size=1000, reduce_ram=True)
	exploration_method=GreedyEpsilon(epsilon_start=1, epsilon_end=0.1, decay=0.001)
	learner=dqn_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
	learner.fit(2)

	assert len(data.x.info)==4
	assert 0 in data.x.info
	assert 3 in data.x.info


def test_export_learner():
	data=MDPDataBunch.from_env('CartPole-v0', render='rgb_array', bs=5, max_steps=20, add_valid=False)
	model=create_dqn_model(data, FixedTargetDQNModule, opt=torch.optim.RMSprop)
	memory=ExperienceReplay(memory_size=1000, reduce_ram=True)
	exploration_method=GreedyEpsilon(epsilon_start=1, epsilon_end=0.1, decay=0.001)
	learner=dqn_learner(data=data, model=model, memory=memory, exploration_method=exploration_method)
	learner.fit(2)

	learner.export('test_export.pkl')#, pickle_data=True)
	learner = load_learner(learner.path, 'test_export.pkl')
	learner.fit(2)
